# AREA A1: the import statements
# import statements needed by esoreflex
import reflex
import sys
from argparse import ArgumentParser
import json
import os
import importlib
import json
import copy
import re

# ------------------------------------------------------------------------------------------
# Import statements needed by the algorithm to execute
from astropy.io import fits as fits
import numpy as np
def do_stuff(files,output_dir):
    if len(files) > 0 :
        hdu = fits.open(files[0].name,method='readonly')
        header = hdu[0].header
        instrument=header['INSTRUME']
        hdu.close()

        try:
            # First try for a user version -- this allows users to "easily" set up support
            # for new instruments placing their python scripts somewhere a 'my_molecfit' directory
            # in a sys.path directory, e.g.:
            # on macOS:
            #   $HOME/Library/Python/3.6/lib/python/site-packages/my_molecfit
            # or on standard Linuxs
            #   $HOME/.local/lib/python3.6/site-packages/my_molecfit
            sys.path.insert(0, "%s/KeplerData/workflows/MyWorkflows/my_molecfit" %(os.environ.get('HOME')))
            inst = importlib.import_module('%s' %(instrument.lower()))
        except ImportError :
            try:
                inst = importlib.import_module('%s' %(instrument.lower()))
            except ImportError :
                # ...but when run in devel mode, the <inst>.py scripts are still in the instruments/ directory
                try:
                    sys.path.insert(0, '%s/instruments' %(os.path.dirname(__file__)))
                    inst = importlib.import_module('%s' %(instrument.lower()))
                except ImportError :
                    raise Exception('%s not (yet?) supported.' % (instrument))
        except Exception as e :
            raise(e)

        try:
            return inst.check_format(files,output_dir)
        except AttributeError as e :
            # If we are here, module exits but has no such method...
            output_files=list()
            for file in files:
                output_files.append(reflex.FitsFile(file.name, file.category, None, file.purposes ))
            return output_files
        except Exception as e :
            raise(e)
    return []
# ------------------------------------------------------------------------------------------
if __name__ == '__main__':

    # ***  PART P1: Input/output ports ***

    #Define inputs/outputs
    parser = reflex.ReflexIOParser()
    parser.add_option("-i", "--in_sof", dest="in_sof")
    parser.add_output("-p", "--out_sof", dest="out_sof")
    inputs = parser.get_inputs()
    outputs = parser.get_outputs()
    in_sof = inputs.in_sof
    files = in_sof.files

    # Define the list of outputs
    output_datasetname=in_sof.datasetName
    # Get the name of the output directory
    pattern = '^--products-dir='
    for arg in sys.argv:
        m=re.match(pattern, arg)
        if m is not None :
            output_dir = re.sub(pattern, '', arg)

    # ***  END OF PART P1 ***
    
    # ***  PART P2: customized part***
    output_files=do_stuff(files,output_dir)
    new_sof = reflex.SetOfFiles(output_datasetname,output_files)
    outputs.out_sof = new_sof
    
    # broadcast outputs:
    parser.write_outputs()
    sys.exit()
    # ***  END OF PART P3 ***
    # ***  END OF AREA A3***    
