# AREA A1: the import statements
# import statements needed by esoreflex
import reflex
import sys
from argparse import ArgumentParser
import json
import os
import re

# Import statements needed by the 
# algorithm to execute
from astropy.io import fits as fits
import numpy as np

# END OF AREA A1

# AREA A2: the algorithm(s).
def normalize(filename,output_):
    hdu = fits.open(filename)
    header = hdu[0].header
    prodcatg=header.get('PRODCATG','NULL')

    if  prodcatg == "SCIENCE.SPECTRUM" :
        # IDP=Binary Table format spectra, i.e.
        # Binary table's in 1 or more HDUs each with
        # WAVE,
        # FLUX,
        # ERR,
        # columns...
        # i.e. this is a bit restrictive and should be improved...
        medians=[]
        for ihdu in hdu[1:] :
            try:
                d=ihdu.data
                dt=d['FLUX']
                try :
                    if len(d['ERR'][~np.isnan(d['ERR'])]) > 0 :
                        dt=d['FLUX'][np.where(d['ERR'][~np.isnan(d['ERR'])] > 0)]
                except :
                    # Not all IDPs have meanignful ERR column, or indeed an ERR column at all...
                    pass
                medians+=[np.median(dt[~np.isnan(dt)]),]
            except:
                continue
        median_cont = float(np.mean(np.array(medians)[~np.isnan(medians)]))
        for ihdu in hdu[1:] :
            ihdu.data['FLUX']=ihdu.data['FLUX']/median_cont  # FLUX column is normalized
            try:
                # Not all IDPs have meanignful ERR column, or indeed an ERR column at all...
                ihdu.data['ERR']=ihdu.data['ERR']/median_cont #ERR  column is normalized
            except :
                pass

    else:
        # non IDP format spectra
        # WORKS FOR IMAGE FORMAT ONLY
        ext=0
        dt=hdu[0].data
        dt=dt[np.where(dt > 0)]
        if dt[0] == None:  # data are not in primary extension
            dt=hdu[1].data
            ext=1
        median_cont=np.median(dt[np.where(~np.isnan(dt))])
        hdu[ext].data=hdu[ext].data/median_cont # DATA extension is normalized
        #Check if there is an extension after data (the error) and normalize that one as well
        try:
            err=hdu[ext+1].data
            hdu[ext+1].data=err/median_cont
        except:
            noerror=1
   
    hdu.writeto(output_, overwrite=True, checksum=True, output_verify='silentfix')
    hdu.close()
# END OF AREA A2
    
if __name__ == '__main__':
    # AREA A3: the interface.
    # This is the main Python script. It contains 2 general parts and a
    # customized part, that is responsible of calling the desired
    # reduction function.

    # ***  PART P1: Input/output ports ***

    #Define inputs/outputs
    parser = reflex.ReflexIOParser()
    parser.add_option("-i", "--in_sof", dest="in_sof")
    parser.add_output("-o", "--out_sof", dest="out_sof")

    inputs = parser.get_inputs()
    outputs = parser.get_outputs()
    in_sof = inputs.in_sof
    files = in_sof.files
    #in_sof2 = inputs.extra_input_port
    #files2 = in_sof2.files

    #Define the list of outputs
    output_files=list()
    output_datasetname=in_sof.datasetName

    #Get the name of the output directory
    pattern = '^--products-dir='
    for arg in sys.argv:
        m=re.match(pattern, arg)
        if m is not None :
            output_dir = re.sub(pattern, '', arg)

    # ***  END OF PART P1 ***

    # ***  PART P2: customized part***

    files_dict={}
    for file in files:
        if file.category not in files_dict.keys() :
            # write the first file of each category to files_dict
            # In the case where there would be multiple SCIENCE
            # molecfit_model would anyway only use the first one...
            files_dict[file.category]=file
    ## molecfit_model only works on a single file, so we only bother normalizing
    ## the one the molecfit_model would select, and it selects in order the first file
    ## it finds in the SOF of STD_MODEL then SCIENCE_CALCTRANS then SCIENCE.
    file_to_normalize=files_dict.get(
        'STD_MODEL',
        files_dict.get(
            'SCIENCE_CALCTRANS',
            files_dict.get(
                'SCIENCE',
                None
            )
        )
    )
    if file_to_normalize is not None :
        for file in files :
            if file == file_to_normalize :
                output_filename=file.name
                infile = file.name.rsplit('/',1)[1]
                output_filename = output_dir+'/'+infile[0:-5]+'_rescaled.fits'
                try:
                    normalize(file.name, output_filename)
                except:
                    # If the normalization fails (e.g., format not recognized),
                    # broadcast the input as it is.
                    output_filename=file.name
                output_files.append(reflex.FitsFile(output_filename, file.category,None,file.purposes))
            else :
                output_files.append(reflex.FitsFile(file.name, file.category, None, file.purposes))
    else :
        raise Exception("No input file with category=STD_MODEL|SCIENCE_CALCTRANS|SCIENCE")
    # *** END OF PART P2***

    # *** PART P3: broadcast products ***
  
    # After the script is completed, the list of products is
    # broadcasted to the output sof.
    new_sof = reflex.SetOfFiles(output_datasetname,output_files)
    outputs.out_sof = new_sof

    # broadcast outputs:
    parser.write_outputs()
    sys.exit()
    # ***  END OF PART P3 ***
    # ***  END OF AREA A3***    
