# AREA A1: the import statements
# import statements needed by esoreflex
import reflex
import sys
from argparse import ArgumentParser
import json
import os
import importlib
import re

# Import statements needed by the 
# algorithm to execute
from astropy.io import fits as fits
import numpy as np

# END OF AREA A1
# AREA A2: the algorithm(s).
# END OF AREA A2
# AREA A3: the interface.
# This is the main Python script. It contains 2 general parts and a
# customized part, that is responsible of calling the desired
# reduction function.
# ------------------------------------------------------------------------------------------
def complete_init():
    return {
        "CALCTRANS_MAPPING_KERNEL": "NULL",
        "CHIP_EXTENSIONS_CALCTRANS": "False",
        "CHIP_EXTENSIONS_CORRECT": "False",
        "CHIP_EXTENSIONS_MODEL": "False",
        "CLEAN_MODEL_FLUX": "False",
        "COLUMN_DFLUX": "ERR",
        "COLUMN_FLUX": "FLUX",
        "COLUMN_LAMBDA": "WAVE",
        "COLUMN_MASK": "NULL",
        "CONST_BARY_RV": 0,
        "CONTINUUM_CONST": 1e-13,
        "CONTINUUM_N": 1,
        "DEFAULT_ERROR": 0.01,
        "ELEVATION_KEYWORD": "ESO TEL GEOELEV",
        "ELEVATION_VALUE": 2635,
        "EMIX": 5,
        "FIT_CONTINUUM": "1",
        "FIT_MOLEC": "NULL",
        "FIT_RES_BOX": "False",
        "FIT_RES_GAUSS": "True",
        "FIT_RES_LORENTZ": "False",
        "FIT_TELESCOPE_BACKGROUND": "False",
        "FIT_WLC": "1",
        "FLUX_UNIT": 0,
        "FTOL": 1e-10,
        "GDAS_PROFILE": "auto",
        "KERNFAC": 3,
        "KERNMODE": "False",
        "LATITUDE_KEYWORD": "ESO TEL GEOLAT",
        "LATITUDE_VALUE": -24.6276,
        "LAYERS": "True",
        "LBLRTM_ALFAL0": 0,
        "LBLRTM_ALTD1": 0,
        "LBLRTM_ALTD2": 0,
        "LBLRTM_AVMASS": 0,
        "LBLRTM_AVTRAT": 2,
        "LBLRTM_BETA": 0,
        "LBLRTM_DELV": 1,
        "LBLRTM_DPTFAC": 0.001,
        "LBLRTM_DPTMIN": 0.0002,
        "LBLRTM_H1": 2.64,
        "LBLRTM_H2": 0,
        "LBLRTM_HOBS": 0,
        "LBLRTM_HSPACE": 120,
        "LBLRTM_IAERSL": 0,
        "LBLRTM_ICNTNM": 5,
        "LBLRTM_IPUNCH": 0,
        "LBLRTM_ITYPE": 3,
        "LBLRTM_LEN": 0,
        "LBLRTM_MODEL": 0,
        "LBLRTM_MPTS": 5,
        "LBLRTM_NOPRNT": 0,
        "LBLRTM_NOZERO": 0,
        "LBLRTM_NPTS": 5,
        "LBLRTM_RANGE": 0,
        "LBLRTM_RE": 0,
        "LBLRTM_REF_LAT": -24.63,
        "LBLRTM_SAMPLE": 4,
        "LBLRTM_SREMIS1": 0,
        "LBLRTM_SREMIS2": 0,
        "LBLRTM_SREMIS3": 0,
        "LBLRTM_SRREFL1": 0,
        "LBLRTM_SRREFL2": 0,
        "LBLRTM_SRREFL3": 0,
        "LBLRTM_TBOUND": 0,
        "LBLRTM_TDIFF1": 5,
        "LBLRTM_TDIFF2": 8,
        "LBLRTM_V1": 1.9,
        "LBLRTM_V2": 2.4,
        "LIST_MOLEC": "NULL",
        "LNFL_LINE_DB": "aer_v_3.8.1.2",
        "LNFL_LINE_DB_FORMAT": 100,
        "LONGITUDE_KEYWORD": "ESO TEL GEOLON",
        "LONGITUDE_VALUE": -70.4051,
        "MAP_REGIONS_TO_CHIP": "1",
        "MAPPING_ATMOSPHERIC": "0,1",
        "MAPPING_CONVOLVE": "1,1",
        "MAPPING_CORRECT": "0,1",
        "MASK_BINARY": "True",
        "MASK_THRESHOLD": 0,
        "MIRROR_TEMPERATURE_KEYWORD": "ESO TEL TH M1 TEMP",
        "MIRROR_TEMPERATURE_VALUE": 15,
        "MODEL_MAPPING_KERNEL": "NULL",
        "OBS_ERF_RV_KEY": "NONE",
        "OBS_ERF_RV_VALUE": 0.,
        "OBSERVING_DATE_KEYWORD": "MJD-OBS",
        "OBSERVING_DATE_VALUE": "-1",
        "OUTPUT_NAME": "out",
        "OUTPUT_PATH": "TELLURICCORR_PARAMETER_DEFAULT",
        "PIXEL_EXCLUDE": "NULL",
        "PIX_SCALE_KEYWORD": "NONE",
        "PIX_SCALE_VALUE": 0.015,
        "PRESSURE_KEYWORD": "ESO TEL AMBI PRES START",
        "PRESSURE_VALUE": 750,
        "PWV": "-1",
        "REFERENCE_ATMOSPHERIC": "equ.fits",
        "RELATIVE_HUMIDITY_KEYWORD": "ESO TEL AMBI RHUM",
        "RELATIVE_HUMIDITY_VALUE": 15,
        "REL_COL": "NULL",
        "RES_BOX": 0,
        "RES_GAUSS": 2.5,
        "RES_LORENTZ": 0,
        "SILENT_EXTERNAL_BINS": "True",
        "SLIT_WIDTH_KEYWORD": "NONE",
        "SLIT_WIDTH_VALUE": 1.0,
        "SUPPRESS_EXTENSION": "False",
        "TELESCOPE_ANGLE_KEYWORD": "ESO TEL ALT",
        "TELESCOPE_ANGLE_VALUE": 90,
        "TELESCOPE_BACKGROUND_CONST": 0.1,
        "TELLURICCORR_DATA_PATH": "TELLURICCORR_PARAMETER_DEFAULT",
        "TELLURICCORR_PATH": "TELLURICCORR_PARAMETER_DEFAULT",
        "TEMPERATURE_KEYWORD": "ESO TEL AMBI TEMP",
        "TEMPERATURE_VALUE": 15,
        "TMP_PATH": "TELLURICCORR_PARAMETER_DEFAULT",
        "TRANSMISSION": "True",
        "USE_DATA_EXTENSION_AS_DFLUX_CALCTRANS": 0,
        "USE_DATA_EXTENSION_AS_DFLUX_CORRECT": 0,
        "USE_DATA_EXTENSION_AS_DFLUX_MODEL": 0,
        "USE_DATA_EXTENSION_AS_MASK_CALCTRANS": 0,
        "USE_DATA_EXTENSION_AS_MASK_CORRECT": 0,
        "USE_DATA_EXTENSION_AS_MASK_MODEL": 0,
        "USE_INPUT_KERNEL_CALCTRANS": "True",
        "USE_INPUT_KERNEL_MODEL": "True",
        "USE_ONLY_INPUT_PRIMARY_DATA_CALCTRANS": "False",
        "USE_ONLY_INPUT_PRIMARY_DATA_CORRECT": "False",
        "USE_ONLY_INPUT_PRIMARY_DATA_MODEL": "False",
        "UTC_KEYWORD": "UTC",
        "UTC_VALUE": "-1",
        "VARKERN": "False",
        "WAVELENGTH_FRAME": "AIR",
        "WAVE_EXCLUDE": "NULL",
        "WAVE_INCLUDE": "NULL",
        "WLC_CONST": 0,
        "WLC_N": 1,
        "WLC_REF": "DATA",
        "WLG_TO_MICRON": 1,
        "XTOL": 1e-10,
    }
# ------------------------------------------------------------------------------------------
def get_inst_setup(filename):
    hdu = fits.open(filename,method='readonly')
    header = hdu[0].header
    instrument = header['INSTRUME']

    try:
        # First try for a user version -- this allows users to "easily" set up support
        # for new instruments placing their python scripts somewhere a 'my_molecfit' directory
        # in a sys.path directory, e.g.:
        # on macOS:
        #   $HOME/Library/Python/3.6/lib/python/site-packages/my_molecfit
        # or on a 'standard' Linux:
        #   $HOME/.local/lib/python3.6/site-packages/my_molecfit
        sys.path.insert(0, "%s/KeplerData/workflows/MyWorkflows/my_molecfit" %(os.environ.get('HOME')))
        inst = importlib.import_module('%s' %(instrument.lower()))
    except ImportError :
        # When installed, the <inst>.py scripts are copied into the same directory
        # as this script...
        try:
            inst = importlib.import_module('%s' %(instrument.lower()))
        except ImportError :
            # ...but when run in devel mode, the <inst>.py scripts are still in the instruments/ directory
            try:
                sys.path.insert(0, '%s/instruments' %(os.path.dirname(__file__)))
                inst = importlib.import_module('%s' %(instrument.lower()))
            except ImportError :
                raise Exception('%s not (yet?) supported.' % (instrument))
    except Exception as e :
        raise(e)

    return (
        inst,
        inst.set_inst_setup(header),
    )

# ------------------------------------------------------------------------------------------
def set_parameters(filename, previous_inst_setup='None', files=None, inst=None):
    hdu = fits.open(filename,method='readonly')
    header = hdu[0].header
    instrument = header.get('INSTRUME','ANY')

    if inst is None :
        try:
            # First try for a user version -- this allows users to "easily" set up support
            # for new instruments placing their python scripts somewhere a 'my_molecfit' directory
            # in a sys.path directory, e.g.:
            # on macOS:
            #   $HOME/Library/Python/3.6/lib/python/site-packages/my_molecfit
            # or on a 'standard' Linux:
            #   $HOME/.local/lib/python3.6/site-packages/my_molecfit
            sys.path.insert(0, "%s/KeplerData/workflows/MyWorkflows/my_molecfit" %(os.environ.get('HOME')))
            inst = importlib.import_module('%s' %(instrument.lower()))
        except ImportError :
            # When installed, the <inst>.py scripts are copied into the same directory
            # as this script...
            try:
                inst = importlib.import_module('%s' %(instrument.lower()))
            except ImportError :
                # ...but when run in devel mode, the <inst>.py scripts are still in the instruments/ directory
                try:
                    sys.path.insert(0, '%s/instruments' %(os.path.dirname(__file__)))
                    inst = importlib.import_module('%s' %(instrument.lower()))
                except ImportError :
                    raise Exception('%s not (yet?) supported.' % (instrument))
        except Exception as e :
            raise(e)

    return (
        inst.set_parameters(
            header=header,
            hdu=hdu,
            previous_inst_setup=inputs.previous_inst_setup,
            files=files,
        )
    )

# ------------------------------------------------------------------------------------------
if __name__ == '__main__':

    # ***  PART P1: Input/output ports ***

    #Define inputs/outputs
    parser = reflex.ReflexIOParser()
    parser.add_option("-i", "--in_sof", dest="in_sof")
    parser.add_option("--current_instrument", dest="current_instrument")
    parser.add_option("--previous_instrument", dest="previous_instrument")
    parser.add_option("--previous_inst_setup", dest="previous_inst_setup")
    parser.add_option("-c", "--complete", dest="complete")
    parser.add_option("-p", "--in_sop", dest="in_sop")

    parser.add_output("-o", "--out_sop", dest="out_sop")
    parser.add_output("--inst_setup", dest="inst_setup")

    inputs = parser.get_inputs()
    outputs = parser.get_outputs()
    in_sof = inputs.in_sof
    files = in_sof.files

    #Define the list of outputs
    output_files=list()
    output_datasetname=in_sof.datasetName

    #Get the name of the output directory
    pattern = '^--products-dir='
    for arg in sys.argv:
        m=re.match(pattern, arg)
        if m is not None :
            output_dir = re.sub(pattern, '', arg)

    # ***  END OF PART P1 ***

    # ***  PART P2: customized part***
    inst_changed=inputs.current_instrument != inputs.previous_instrument

    c_attrs={}
    o_attrs={}
    inst_setup=None
    for file in files:
        if file.category in ['SCIENCE','SCIENCE_CALCTRANS','STD_MODEL',]:
            if inst_setup is None :
                inst, inst_setup=get_inst_setup(
                    file.name,
                )
                if inputs.complete == "True" :
                    c_attrs = complete_init()
                    inst_setup="None"
                else :
                    # force a full setup each time now that we have save_params working...
                    c_attrs = complete_init()
                    o_attrs = set_parameters(
                        file.name,
                        previous_inst_setup=inputs.previous_inst_setup,
                        files=files,
                        inst=inst,
                    )

        output_files.append(reflex.FitsFile(file.name, file.category,None,file.purposes))

    # *** END OF PART P2***

    # *** PART P3: broadcast products ***
  
    # After the script is completed, the list of products is
    # broadcasted to the output sof.
    new_sof = reflex.SetOfFiles(output_datasetname,output_files)
    outputs.out_sof = new_sof
    # Create the output SOP...
    out_sop=[]
    sop_keynames=[rp.name for rp in inputs.in_sop]
    out_sop_keys={}
    for _d in [c_attrs,o_attrs] :
        for k in _d.keys() :
            if _d[k] is not None :
                if k in sop_keynames :
                    if inputs.in_sop[sop_keynames.index(k)].valtype == 'bool' :
                        if isinstance(_d[k],str) :
                            _d[k]=_d[k].lower() == 'true'
                    if _d[k] != inputs.in_sop[sop_keynames.index(k)].value :
                        if k in out_sop_keys.keys() :
                            out_sop[out_sop_keys[k]].value=_d[k]
                        else :
                            out_sop+=[inputs.in_sop[sop_keynames.index(k)],]
                            out_sop[-1].value=_d[k]
                            out_sop_keys[k]=len(out_sop)-1
    
    outputs.out_sop=out_sop
    outputs.inst_setup=inst_setup or 'None'

    #  Broadcast outputs on the extra output port, if needed
    #outputs.extra_output_port = new_sof

    # broadcast outputs:
    parser.write_outputs()
    sys.exit()
    # ***  END OF PART P3 ***
    # ***  END OF AREA A3***    
# ------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
'''
## ALL recipe variables in the molecfit wkf:
CALCTRANS_MAPPING_KERNEL
CHIP_EXTENSIONS                      (_MODEL|_CALCTRANS|_CORRECT)
CLEAN_MODEL_FLUX
COLUMN_DFLUX
COLUMN_FLUX
COLUMN_LAMBDA
COLUMN_MASK
CONST_BARY_RV
CONTINUUM_CONST
CONTINUUM_N
DEFAULT_ERROR
ELEVATION_KEYWORD
ELEVATION_VALUE
EMIX
FIT_CHIP1                            **not implemented in wkf**
FIT_CHIP2                            **not implemented in wkf**
FIT_CHIP3                            **not implemented in wkf**
FIT_CHIP4                            **not implemented in wkf**
FIT_CONTINUUM
FIT_MOLEC
FIT_RANGE1                            **not implemented in wkf**
FIT_RANGE2                            **not implemented in wkf**
FIT_RANGE3                            **not implemented in wkf**
FIT_RANGE4                            **not implemented in wkf**
FIT_RES_BOX
FIT_RES_GAUSS
FIT_RES_LORENTZ
FIT_TELESCOPE_BACKGROUND
FIT_WLC
FLUX_UNIT
FTOL
GDAS_PROFILE
KERNFAC
KERNMODE
LATITUDE_KEYWORD
LATITUDE_VALUE
LAYERS
LBLRTM_ALFAL0
LBLRTM_ALTD1
LBLRTM_ALTD2
LBLRTM_AVMASS
LBLRTM_AVTRAT
LBLRTM_BETA
LBLRTM_DELV
LBLRTM_DPTFAC
LBLRTM_DPTMIN
LBLRTM_H1
LBLRTM_H2
LBLRTM_HOBS
LBLRTM_HSPACE
LBLRTM_IAERSL
LBLRTM_ICNTNM
LBLRTM_IPUNCH
LBLRTM_ITYPE
LBLRTM_LEN
LBLRTM_MODEL
LBLRTM_MPTS
LBLRTM_NOPRNT
LBLRTM_NOZERO
LBLRTM_NPTS
LBLRTM_RANGE
LBLRTM_RE
LBLRTM_REF_LAT
LBLRTM_SAMPLE
LBLRTM_SREMIS1
LBLRTM_SREMIS2
LBLRTM_SREMIS3
LBLRTM_SRREFL1
LBLRTM_SRREFL2
LBLRTM_SRREFL3
LBLRTM_TBOUND
LBLRTM_TDIFF1
LBLRTM_TDIFF2
LBLRTM_V1
LBLRTM_V2
LIST_MOLEC
LNFL_LINE_DB
LNFL_LINE_DB_FORMAT
LONGITUDE_KEYWORD
LONGITUDE_VALUE
MAPPING_ATMOSPHERIC
MAPPING_CONVOLVE
MAPPING_CORRECT
MASK_BINARY
MASK_THRESHOLD
MIRROR_TEMPERATURE_KEYWORD
MIRROR_TEMPERATURE_VALUE
MODEL_MAPPING_KERNEL
OBSERVING_DATE_KEYWORD
OBSERVING_DATE_VALUE
OBS_ERF_RV_KEY
OBS_ERF_RV_VALUE
OUTPUT_NAME
OUTPUT_PATH
PIXEL_EXCLUDE
PIX_SCALE_KEYWORD
PIX_SCALE_VALUE
PRESSURE_KEYWORD
PRESSURE_VALUE
PWV
REFERENCE_ATMOSPHERIC
RELATIVE_HUMIDITY_KEYWORD
RELATIVE_HUMIDITY_VALUE
REL_COL
RES_BOX
RES_GAUSS
RES_LORENTZ
SILENT_EXTERNAL_BINS
SLIT_WIDTH_KEYWORD
SLIT_WIDTH_VALUE
SUPPRESS_EXTENSION
TELESCOPE_ANGLE_KEYWORD
TELESCOPE_ANGLE_VALUE
TELESCOPE_BACKGROUND_CONST
TELLURICCORR_DATA_PATH
TELLURICCORR_PATH
TEMPERATURE_KEYWORD
TEMPERATURE_VALUE
TMP_PATH
TRANSMISSION
USE_DATA_EXTENSION_AS_DFLUX                     (_MODEL|_CALCTRANS|_CORRECT)
USE_DATA_EXTENSION_AS_MASK                      (_MODEL|_CALCTRANS|_CORRECT)
USE_INPUT_KERNEL                                (_MODEL|_CALCTRANS)
USE_ONLY_INPUT_PRIMARY_DATA                     (_MODEL|_CALCTRANS|_CORRECT)
UTC_KEYWORD
UTC_VALUE
VARKERN
WAVELENGTH_FRAME
WAVE_EXCLUDE
WAVE_INCLUDE
WLC_CONST
WLC_N
WLC_REF
WLG_TO_MICRON
XTOL
'''
