from sys import dont_write_bytecode
import reflex
import numpy as np
from astropy.io import fits
from astropy.table import Table
import re
import os
import sys
import importlib
import copy

try :
    am = importlib.import_module('auto_molecule')
    iu = importlib.import_module('inst_utils')
except ImportError:
    sys.path.insert(0, '%s/instruments' %(os.path.dirname(__file__)))
    am = importlib.import_module('auto_molecule')
    iu = importlib.import_module('inst_utils')
except Exception as e:
    raise(e)

UV_slit_width_keywords={
    "BLUE": "ESO INS SLIT2 WID",
    "RED":  "ESO INS SLIT3 WID",
    "UNKNOWN": "NONE",
}

# ------------------------------------------------------------------------------------------
def set_inst_setup(header=None) :
    if header :
        return "%s_%s" %(
            re.sub('^.*_','',header.get('HIERARCH ESO PRO CATG', 'UNKNOWN')),
            header.get('HIERARCH ESO INS GRAT1 WLEN', header.get('HIERARCH ESO INS GRAT2 WLEN', 'UNKNOWN')),
        )
    return 'None'
# ------------------------------------------------------------------------------------------
def check_format(files, output_dir):
    output_files=list()
    files_dict={}
    science_dict={}
    err_dict={}
    for file in files:
        # OCA does not actually suport STD_MODEL yet...
        if file.category in ['STD_MODEL','SCIENCE','SCIENCE_CALCTRANS',] :
            if file.category not in science_dict.keys() :
                science_dict[file.category]=[]
            science_dict[file.category]+=[file]
        elif file.category in ['STD_ERRORS','SCIENCE_ERRORS',] :
            if file.category not in err_dict.keys() :
                err_dict[file.category]=[]
            err_dict[file.category]+=[file]
        else :
            files_dict[file.category]=file
    has_sci_and_std=(
        ('SCIENCE' in files_dict.keys() or 'SCIENCE_CALCTRANS' in files_dict.keys())
        and
        'STD_MODEL' in files_dict.keys()
    )
    for k in science_dict.keys() :
        if len(science_dict[k]) > 1 :
            if len(science_dict[k]) == 2 :
                err_file_names={
                    'FLUXCAL_SCI_POINT_RED': 'FLUXCAL_ERRORBAR_SCI_POINT_',
                    'RED_SCI_POINT_RED': 'ERRORBAR_SCI_POINT_',
                }
                # assume these are the 'traditional' REDL + REDU files produced by the ESO PL,
                # combine into a single binary-table
                wave_flux_err={}
                naxes={}
                for file in science_dict[k] :
                    hdulist=fits.open(file.name)
                    chip=re.sub('^.*_','',hdulist[0].header.get('HIERARCH ESO PRO CATG', 'UNKNOWN'))
                    new_procatg=re.sub('_RED.$','_RED',hdulist[0].header.get('HIERARCH ESO PRO CATG','UNKNOWN'))
                    wave_flux_err[chip]={
                        'wave': (
                            (np.arange(1, hdulist[0].header['NAXIS1']+1, 1)-hdulist[0].header['CRPIX1'])
                            *
                            hdulist[0].header.get('CD1_1',hdulist[0].header.get('CDELT1')) + hdulist[0].header['CRVAL1']
                        ),
                        'flux': hdulist[0].data,
                    }
                    naxes[chip]={
                        'NAXIS1': hdulist[0].header['NAXIS1'],
                        'CRPIX1': hdulist[0].header['CRPIX1'],
                        'CD1_1': hdulist[0].header['CD1_1'],
                        'CDELT1': hdulist[0].header['CDELT1'],
                        'CRVAL1': hdulist[0].header['CRVAL1'],
                        'CRVALN': wave_flux_err[chip]['wave'][-1],
                    }
                    hdr=copy.deepcopy(hdulist[0].header)
                    hdulist.close()
                    # broadcast the original frame (for sticking it all back together at the end...)
                    output_files.append(reflex.FitsFile(
                        file.name, "ORIG_%s" %(file.category), None, file.purposes
                    ))
                for err_k in err_dict.keys() :
                    for err_file in err_dict[err_k] :
                        err_hdulist=fits.open(err_file.name)
                        chip=re.sub('^.*_','',err_hdulist[0].header.get('HIERARCH ESO PRO CATG', 'UNKNOWN'))
                        if err_file_names[new_procatg] in err_hdulist[0].header.get('HIERARCH ESO PRO CATG', 'UNKNOWN') :
                            wave_flux_err[chip]['err']= err_hdulist[0].data
                        err_hdulist.close()
                _err={}
                for chip in wave_flux_err :
                    _err[chip]=wave_flux_err[chip].get('err',np.ones(len(wave_flux_err[chip]['wave'])))
                    # ------------------------------------------------------------------
                    ## Apparently the following is no-longer true since version ?.?.?
                    # But keep this for legacy...
                    if 'err' in wave_flux_err[chip] and new_procatg in ['RED_SCI_POINT_RED',] :
                        # 'RED_SCI_POINT_RED' files have err = 1.0 for pixels that were
                        # not actually extracted
                        # where as 'FLUXCAL_SCI_POINT_RED' files have err = 0.0 for pixels that were
                        # not actually extracted
                        wave_flux_err[chip]['err'][
                            np.logical_and(
                                wave_flux_err[chip]['err'] == 1.0,
                                wave_flux_err[chip]['flux'] == 0.0,
                            )
                        ]=0.
                    # ------------------------------------------------------------------
                wave=np.array([])
                flux=np.array([])
                err=np.array([])
                include_err=True
                for chip in ['REDL','REDU',] :
                    wave=np.append(wave,wave_flux_err[chip]['wave'][_err[chip] > 0.])
                    flux=np.append(flux,wave_flux_err[chip]['flux'][_err[chip] > 0.])
                    if 'err' in wave_flux_err[chip] :
                        err=np.append(err,wave_flux_err[chip]['err'][_err[chip] > 0.])
                    else :
                        include_err=False
                        #err=np.append(err,wave_flux_err[chip]['flux']*0.+1.e-4*np.max(wave_flux_err[chip]['flux']))
                hdr['PRODCATG']='SCIENCE.SPECTRUM'
                orig_catg=hdr['HIERARCH ESO PRO CATG']
                hdr['HIERARCH ESO PRO CATG']=new_procatg
                hdr['PRODCATG']="SCIENCE.SPECTRUM"
                # extra kws to make files IDP compliant enough to satisfy
                # esotk_spectrum1d_combine recipe
                hdr['TEXPTIME']=hdr['EXPTIME']
                hdr['MJD-END']=hdr['MJD-OBS']-hdr['EXPTIME']/24./60./60.
                hdr['NCOMBINE']=1
                hdr['OBID']=hdr['HIERARCH ESO OBS ID']
                hdr['PROV']=hdr['HIERARCH ESO PRO REC1 RAW1 NAME']
                hdr['TOT_FLUX']=0.
                hdr['SPEC_RES']=0.

                fmt=str(len(wave))+'D'
                c1=fits.Column(name='WAVE', array=[wave], format=fmt, unit=hdulist[0].header['CUNIT1'])
                # wave is in air for UVES PL products
                # Transform to topo_vac for molecfit_model
                wave_tv=iu.air_to_vac(wave,wlg)
                c4=fits.Column(name='WAVE_TV', array=[wave_tv], format=fmt, unit=hdulist[0].header['CUNIT1'])
                c2=fits.Column(name='FLUX', array=[flux], format=fmt, unit=hdulist[0].header['BUNIT'])
                cols=[c1,c4,c2]
                if include_err :
                    c3=fits.Column(name='ERR', array=[err], format=fmt, unit=hdulist[0].header['BUNIT'])
                    cols+=[c3,]
                h0=fits.PrimaryHDU(data=None,header=hdr)

                h1 = fits.BinTableHDU.from_columns(cols)
                h1.header['EXTNAME']='SPECTRUM'
                h1.header['NELEM']=len(wave)
                h1.header['NELEM_L']=len(wave_flux_err['REDL']['wave'][_err['REDL'] > 0.])
                h1.header['NELEM_U']=len(wave_flux_err['REDU']['wave'][_err['REDU'] > 0.])
                h1.header['OBJECT']=hdr['OBJECT']
                h1.header['RA']=hdr['RA']
                h1.header['DEC']=hdr['DEC']
                h1.header['APERTURE']=0.1
                h1.header['HIERARCH ESO MF WAVE COLUMN']='WAVE_TV'
                h1.header['HIERARCH ESO MF WAVE COLUMN WAVELENGTH_FRAME']='TOPO+VAC'
                h1.header['TUCD1']="em.wl;obs.atmos"
                h1.header['TUTYP1']="spec:Data.SpectralAxis.Value"
                h1.header['TUNIT1']="nm"
                h1.header['TUCD2']="em.wl;obs.vacuum"
                h1.header['TUTYP2']="spec:Data.SpectralAxis.Value"
                h1.header['TUNIT2']="nm"
                h1.header['TUCD3']="phot.flux.density;em.wl;src.net;meta.main"
                h1.header['TUTY3']="spec:Data.FluxAxis.Value"
                h1.header['TUNIT3']="adu"
                if include_err :
                    h1.header['TUCD4']="stat.error;phot.flux.density;meta.main"
                    h1.header['TUTYP4']="spec:Data.FluxAxis.Accuracy.StatError"
                    h1.header['TUNIT4']="adu"
                for chip in ['REDL','REDU',] :
                    for k in naxes[chip] :
                        h1.header['%s_%s' %(k,chip[3])]=naxes[chip][k]
                    h1.header['WL_MIN_%s' %(chip[3])]=np.nanmin(wave_flux_err[chip]['wave'][_err[chip] > 0.])
                    h1.header['WL_MAX_%s' %(chip[3])]=np.nanmax(wave_flux_err[chip]['wave'][_err[chip] > 0.])
                h1.header['IND_GAP']=len(wave_flux_err['REDL']['wave'][_err['REDL'] > 0.])
                hl=fits.HDUList([h0,h1])
                new_filename="%s/%s" %(
                    output_dir,
                    os.path.basename(re.sub('_RED.\.fits$','_RED.fits',file.name)),
                )
                print('writing %s' %(new_filename))
                iu.HDUList(hl).writeto(new_filename, overwrite=True, checksum=True, output_verify='silentfix')
                output_files.append(reflex.FitsFile(new_filename, file.category, None, file.purposes ))
            elif len(science_dict[k]) > 2 :
                raise Exception('Can not handle datasets with more than 2 %s files' %(k))
        else :
            file=science_dict[k][0]
            output_files.append(reflex.FitsFile(file.name, file.category, None, file.purposes ))

    for k in files_dict.keys() :
        file=files_dict[k]
        output_files.append(reflex.FitsFile(file.name, file.category, None, file.purposes ))

    return output_files
# ------------------------------------------------------------------------------------------
def recombine_idps( files=None, orig_files=None ) :
    new_files=None
    propagate_keys_file=None
    files_dict={}
    science_dict={}
    err_dict={}
    for file in orig_files:
        # OCA does not actually suport STD_MODEL yet...
        if file.category in ['ORIG_STD_MODEL','ORIG_SCIENCE','ORIG_SCIENCE_CALCTRANS',] :
            if file.category not in science_dict.keys() :
                science_dict[file.category]=[]
            science_dict[file.category]+=[file]
        elif file.category in ['STD_ERRORS','SCIENCE_ERRORS',] :
            if file.category not in err_dict.keys() :
                err_dict[file.category]=[]
            err_dict[file.category]+=[file]
        else :
            files_dict[file.category]=file
    for file in files:
        files_dict[file.category]=file
        if file.category in ["SCIENCE_TELLURIC_CORR",] :
            propagate_keys_file=propagate_keys_file or file

    output_dir=os.path.dirname(propagate_keys_file.name or files[0].name)
    k='SCIENCE_TELLURIC_CORR'
    if k in files_dict.keys() :
        for sk in science_dict.keys() :
            corrected_hdulist=fits.open(files_dict['SCIENCE_TELLURIC_CORR'].name)
            indexes={
                'REDL': [0,corrected_hdulist[1].header['NELEM_L']],
                'REDU': [corrected_hdulist[1].header['NELEM_L']+1,-1],
            }
            if len(science_dict[sk]) > 1 :
                if len(science_dict[sk]) == 2 :
                    # assume these are the 'traditional' REDL + REDU files produced by the ESO PL,
                    # combine into a single binary-table
                    for file in science_dict[sk] :
                        new_files=new_files or []
                        hdr=None
                        if propagate_keys_file is not None :
                            hdr=iu.propagate_keys(
                                propagate_keys_file,
                                ["molecfit_model","molecfit_calctrans","molecfit_correct"],
                                orig_file=file,
                            )
                        bname=os.path.basename(file.name)
                        hdulist=fits.open(file.name)
                        chip=re.sub('^.*_','',hdulist[0].header.get('HIERARCH ESO PRO CATG', 'UNKNOWN'))
                        hdulist[0].data=corrected_hdulist[1].data['FLUX'][0][indexes[chip][0]:indexes[chip][1]]
                        hdulist[0].header['PIPEFILE']="%s_%s" %(k,bname)
                        for hk in ['HIERARCH ESO PRO CATG',] :
                            hdulist[0].header[hk]='%s_%s' %(
                                corrected_hdulist[0].header[hk],
                                chip,
                            )
                        for hk in ['DATE',] :
                            hdulist[0].header[hk]=corrected_hdulist[0].header[hk]

                        out_file_name="%s/%s_%s" %(output_dir,k,bname)
                        if hdr is not None :
                            hdulist[0].header+=hdr
                        print("writing '%s'..." %out_file_name)
                        hdulist[0].header['PIPEFILE']="%s_%s" %(k,bname)
                        iu.HDUList(hdulist).writeto(
                            out_file_name, overwrite=True, checksum=True, output_verify='silentfix',
                        )
                        print("wrote '%s'." %out_file_name)
                        new_files+=[
                            reflex.FitsFile(
                                out_file_name,
                                k,
                                hdulist[0].header['CHECKSUM'],
                                ['UNIVERSAL',]
                            )
                        ]

    return new_files or files

# ------------------------------------------------------------------------------------------
wlg = 1.0e-04   # WLG_TO_MICRON=0.0001
def set_parameters(header=None, hdu=None, previous_inst_setup='None', files=None) :
    wframe='AIR'
    list_mol='NULL'
    fit_mol='NULL'
    rel_col='NULL'
    clam='WAVE'
    cflux='FLUX'
    cdflux='ERR'
    fit_tel_back="False"
    cont_poly = int(3)
    varkern="FALSE"
    wave_include='NULL'

    return {
        'WLG_TO_MICRON': wlg,
        'WAVELENGTH_FRAME': wframe,
        'LIST_MOLEC': list_mol,
        'FIT_MOLEC': fit_mol,
        'REL_COL': rel_col,
        'COLUMN_LAMBDA': clam,
        'COLUMN_FLUX': cflux,
        'COLUMN_DFLUX': cdflux,
        'FIT_TELESCOPE_BACKGROUND': fit_tel_back,
        'CONTINUUM_N': cont_poly,
        'VARKERN': varkern,
        'WAVE_INCLUDE':wave_include,
        'SLIT_WIDTH_KEYWORD': UV_slit_width_keywords[hdu[0].header.get("ESO INS PATH","UNKNOWN")],
    }
# ------------------------------------------------------------------------------------------
def check_idp(hdu, previous_inst_setup='None', files=None, ParameterInitialization=False, recipe=None, ) :
    import numpy as np
    no_output=0 

    # these are the values to use for UVES IDP spectra
    clam='WAVE'
    cflux='FLUX'
    cdflux='ERR'
    primary='FALSE'
    errorext=0
    maskext=0
    map_convolve="1,1"
    map_atmosphere="0,1"
    map_correct="0,1"

    wl_exclude_gaps=[]
    if hdu[0].header.get('PRODCATG','NULL') in [
      "SCIENCE.SPECTRUM",
    ] or hdu[0].header.get('HIERARCH ESO PRO CATG','NULL') in [
      "FLUXCAL_SCI_POINT_RED",  # IDP
      "RED_SCI_POINT_RED",  # IDP
    ] :
        # these are the values to use for UVES IDP format spectra
        for suffix in ['','_REDUCED'] :
            if 'FLUX%s' %(suffix) in hdu[1].data.columns.names :
                break
        cflux='FLUX%s' %(suffix)
        cdflux='ERR%s' %(suffix)
        dd=hdu[1].data[cflux]
        wl_range=np.array([np.min(hdu[1].data[clam][0]),np.max(hdu[1].data[clam][0])])
        wlR=wl_range[1]-wl_range[0]
        wl_resolution=np.min(hdu[1].data[clam][0][1:]-hdu[1].data[clam][0][0:-1])
        # Check if the 'ERR' column is actually there...
        _t=Table.read(hdu, format='fits')
        # Exclude the WL range between the two chips...
        # determine the bordering WLs and +/- 0.5% of the full WL range...
        exclude_fraction=5.e-03
        index_gap=(
            hdu[1].header.get('IND_GAP',None)
            or
            np.argsort(hdu[1].data[clam][0][1:]-hdu[1].data[clam][0][0:-1])[-1]
        )
        if cdflux in _t.columns :
            # use the fact that err = 0. for non-extracted pixels to workout
            # the full range of the WL gap between chips
            index_last_non_zero_err_before_gap=np.where(hdu[1].data[cdflux][0][0:index_gap] > 0. )[0][-1]
            index_first_non_zero_err_after_gap=index_gap+1+np.where(hdu[1].data[cdflux][0][index_gap+1:] > 0. )[0][0]
        else :
            cdflux="NULL"
            # Don't have err's to reliable determine which were no extracted pixels
            # so just use the index_gap as an estimation...
            index_last_non_zero_err_before_gap=index_gap
            index_first_non_zero_err_after_gap=index_gap+1
        wl_exclude_gaps=[   # a list...
            [  # of lists (one per extension/chip)...
                [   # of lists of the gap(s)
                    hdu[1].data[clam][0][index_last_non_zero_err_before_gap]-wlR*exclude_fraction,
                    hdu[1].data[clam][0][index_first_non_zero_err_after_gap]+wlR*exclude_fraction,
                ],
            ],
        ]
        # For UVES wl_exclude_gaps is [[[gap_start,gap_end],],] but other instruments can be more
        # complicated (e.g. CRIRES)
    else:
        #these are the values to use for UVES spectra in generic PL format (nom IDP)
        primary='TRUE'
        errorext=0
        maskext=0
        map_convolve="1"
        map_atmosphere="1"
        map_correct="1"
        try:
            dd=hdu[0].data
        except:         
            dd=hdu[1].data
        wl_range=np.array(
            [
                hdu[0].header['CRVAL1'],
                hdu[0].header['CRVAL1']+hdu[0].header['NAXIS1']*hdu[0].header.get('CD1_1',hdu[0].header.get('CDELT1'))
            ]
        )
        wl_resolution=hdu[0].header.get('CD1_1',hdu[0].header.get('CDELT1'))
    median_cont = float(np.median(dd[np.where(~np.isnan(dd))]))

    WL_inc_exc_PIX_exc={}
    inst_setup_changed = set_inst_setup(hdu[0].header) != previous_inst_setup
    if ParameterInitialization :
        ## ---------------------------------------------------------------------
        # Automatically set molecules and ranges...
        wl_span=wl_range[1]-wl_range[0]

        WL_ranges_to_include=am.auto_molecules_set_WL_ranges_to_include(
            files,
            wl_range,
            wl_span,
            wl_resolution,
            wl_scale=1./wlg,
            buff=wl_span*0.05,
            wl_exclude_gaps=wl_exclude_gaps,
        )
        WL_inc_exc_PIX_exc={
            'LIST_MOLEC': WL_ranges_to_include['list_mol'],
            'FIT_MOLEC': WL_ranges_to_include['fit_mol'],
            'REL_COL': WL_ranges_to_include['rel_col'],
            'WAVE_INCLUDE': WL_ranges_to_include['wave_include'],
            'WAVE_EXCLUDE': WL_ranges_to_include['wave_exclude'],
            'PIXEL_EXCLUDE': WL_ranges_to_include['pixel_exclude'],
        }
        ## ---------------------------------------------------------------------
    return dict(
        list(({
            # model settings...
            'COLUMN_LAMBDA': clam,
            'COLUMN_FLUX': cflux,
            'COLUMN_DFLUX': cdflux,
            'USE_ONLY_INPUT_PRIMARY_DATA': primary,
            'USE_DATA_EXTENSION_AS_DFLUX': errorext,
            'USE_DATA_EXTENSION_AS_MASK': maskext,
            'CONTINUUM_CONST': median_cont,
            'SLIT_WIDTH_KEYWORD': UV_slit_width_keywords[hdu[0].header.get("ESO INS PATH","UNKNOWN")],
            # calctrans settings...
            'MAPPING_CONVOLVE': map_convolve,
            'MAPPING_ATMOSPHERIC': map_atmosphere,
            # correct settings...
            'MAPPING_CORRECT': map_correct,
            # Only set the following if the canvas value is NULL
            'ignore_if_not_NULL': {} if inst_setup_changed else WL_inc_exc_PIX_exc,
        }).items())
        +
        list((WL_inc_exc_PIX_exc if inst_setup_changed else {}).items())
    )
# ------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
# -----   UVES specific methods   ----------------------------------------------------------
# ------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
